# Superconducting nanostrucutre

# In the case of the superconducting nanostructure, we use a fact, that the Hamiltonian (1) contains 3 harmonic oscillators allowing us to introduce CAPs as $\hat{\phi}_n-\phi_n^r = (\hat{a}_n{}^\dagger + \hat{a}_n)/\alpha_n$, yielding
# $$ \hat{H} = \sum_n \hbar \omega_k \left(\hat{a}_n{}^\dagger\hat{a}_n+\frac{1}{2}\right) + \sum_n\frac{\hbar I_n}{2e}\hat{\phi}_n\hat{\sigma}_n\ $$
# where $\omega_n = 1/\sqrt{L_n C_n}$, $\alpha_n = e\sqrt{2/\hbar}(L_n/C_n)^{1/4}$, and $[\hat{a}_n,\hat{a}_m{}^\dagger] = \delta_{nm}$.

# Since computers only crunch numbers, we frame our problem in dimensionless form as follows:
# $$ \hat{H} = \sum_n \omega_k \left(\hat{a}_n{}^\dagger\hat{a}_n+\frac{1}{2}\right) + \sum_n I_n(\hat a_n^\dagger + \hat a_n + \phi_{0n})\hat{\sigma}_n\ $$
# where one goes back to a dimensional form by substitutions $\omega_k\to\hbar \omega_k$, $I_n \to \hbar I_n/2e\alpha_n$, $\phi_{0n} \to \alpha_n \phi^r_n$.

# ## Generation of Hamiltonian
# To represent elements in our hilbert space to an element in a array we choose a following mapping
# $$ s =  2(N_y+1)(N_z+1)l+2(N_z+1)m+2n+a$$
# where $N_{x,y,z}$ is the cutoff in oscilator space and $l,m,n$ represents a specific harmonic oscilator state $|lmn \pm\rangle$. Here we give the simplest implementation (which can be improved by an optimization for only running over meaningfull indices).  

### The filling of the operator in this mapping then looks as follows
using LinearAlgebra
using SparseArrays
using Arpack

function FillOperator(Op::Function,Nx::Int,Ny::Int,Nz::Int)

    Opm = Array{Complex}(undef,0)
    Iv = Array{Int}(undef,0)
    Jv = Array{Int}(undef,0)
    
    s(l,m,n,a) = 2*(Ny+1)*(Nz+1)*l+2*(Nz+1)*m+2*n+a
    for l in 0:Nx    
        for m in 0:Ny
            for n in 0:Nz
                for a in 1:2
                    for l_ in 0:Nx    
                        for m_ in 0:Ny
                            for n_ in 0:Nz
                                for a_ in 1:2
                                    sbra = s(l_,m_,n_,a_)
                                    sket = s(l,m,n,a)
                                    
                                    Opi = Op(l_,m_,n_,a_,l,m,n,a)
                                    if Opi!=0 || sbra==sket
                                        push!(Iv,sbra)
                                        push!(Jv,sket)
                                        push!(Opm,Opi)
                                    end
                                end
                            end
                        end
                    end
                end
            end
        end
    end
    return sparse(Iv,Jv,Opm)
end

### Our junction Hamiltonian is 
struct ParametersJ
    hw::Tuple{Float64,Float64,Float64}
    I::Tuple{Float64,Float64,Float64}
    phi::Tuple{Float64,Float64,Float64}
end

#divf(x,y) = y==0 ? 0 : x/y

const sigmax = [0 1;1 0] 
const sigmay = [0 -1im;1im 0]
const sigmaz = [1 0;0 -1] 

δ(m,n) = m==n ? 1 : 0

function H(l_,m_,n_,a_,l,m,n,a,p::ParametersJ)

    hw,I,phi= p.hw,p.I,p.phi

    sigma = (sigmax[a_,a],sigmay[a_,a],sigmaz[a_,a]) 
    bc = (sqrt(l+1)*δ(l_,l+1), sqrt(m+1)*δ(m_,m+1), sqrt(n+1)*δ(n_,n+1))
    ba = (sqrt(l)*δ(l_,l-1),sqrt(m)*δ(m_,m-1),sqrt(n)*δ(n_,n-1))

    E = δ(l_,l)*δ(m_,m)*δ(n_,n)

    s::Complex{Float64} = 0

    s += sum(hw.*(l+1/2,m+1/2,n+1/2).*E.*δ(a_,a))
    s += sum(sigma.*I.*(bc.+ba).*(δ(m_,m)*δ(n_,n),δ(l_,l)*δ(n_,n),δ(l_,l)*δ(m_,m)))
    s += sum(sigma.*I.*phi)*E
    return s
end


function energies(hw,I,phi,N,NEV)
    println("phi = $phi")
    #run(`hostname`)
    BLAS.set_num_threads(1)
    p = ParametersJ(hw,I,phi)

    Hm = FillOperator((l_,m_,n_,a_,l,m,n,a)-> abs(l-l_)+abs(n-n_)+abs(m-m_) > 1 ? 0 : H(l_,m_,n_,a_,l,m,n,a,p),N...)
    solution = eigs(Hm,nev=NEV,tol=1e-18,maxiter=1000,which=:SR)

    # Calculation of spin
    spinx = []
    spiny = []
    spinz = []
    for i in 1:length(solution[1])
        up=solution[2][:,i][1:2:end]
        down=solution[2][:,i][2:2:end]

        spinxi = real(up'*down + down'*up)
        spinyi = real(-1im*up'*down + 1im*down'*up)
        spinzi = up'*up - down'*down

        push!(spinx,spinxi)
        push!(spiny,spinyi)
        push!(spinz,spinzi)
    end

    return Dict(:EVALS=>real(solution[1]),:spinx=>spinx,:spiny=>spiny,:spinz=>spinz)
end


function test()
    p = ParametersJ((1,1,1),(1,1,1),(1,0,0))
    N = (15,15,15)
    NEV = 20 ### Number of eigenvalues from the bottom

    ### There is 10X speedup in sparse matrix construction when 0 can be calculated immeadiatelly
    @time Hm = FillOperator((l_,m_,n_,a_,l,m,n,a)-> abs(l-l_)+abs(n-n_)+abs(m-m_) > 1 ? 0 : H(l_,m_,n_,a_,l,m,n,a,p),N...)
    @time solution = eigs(Hm,nev=NEV,tol=1e-18,maxiter=1000,which=:SR)
end
