# The Hamiltonian considered for the Weyl exciton is
# $$
#  \hat{H}_{\rm ex} = \Delta_{\rm ex}+ v_{na}\hat{p}_n\hat{\sigma}_a + \frac{(\hat{p}_n - p^T_n)^2}{2 m_n^*} - \frac{e^2}{4\pi\epsilon_0 \epsilon r}\,,
# $$
# which we put into a simpler form for computer:
# $$
#  \hat{H}_{\rm ex} = \sum_nv_{n}(\partial_n +p_n^T)\hat{\sigma}_n + \frac{\partial_n^2}{2 m_n^*} - \frac{\alpha}{ r}\,,
# $$
# where we used $ \hat{p}_n - p^T_n = - i \partial_n$ and made a choice to count energy from the continium. Becasue of singularity at the center, the convergence on a uniform grid is unsatisfactory. Therefore we choose a logarithmic grid which for enumerated points $1...N$: 
# $$
# x(i,N,L) =   \begin{cases}
#    -(L+1)^{(N/2-i+1)/(N/2+1)} + \frac{1}{2}(1 + (L+1)^{1/(N/2 + 1)})     & \quad \text{if } i\leq N/2\\
#    (L+1)^{(i-N/2)/(N/2+1)} - \frac{1}{2}(1 + (L+1)^{1/(N/2 + 1)})  & \quad \text{if } i > N/2
#   \end{cases}
# $$

# Also we found that convergence with respect to a box size $L$ was unsatisfactory thus we used Richardson extrapolation to improve the convergence. 

using LinearAlgebra
using SparseArrays
using Arpack

function FillOperator(Op::Function,Nx::Int,Ny::Int,Nz::Int)

    Opm = Array{Complex}(undef,0)
    Iv = Array{Int}(undef,0)
    Jv = Array{Int}(undef,0)
    
    s(l,m,n,a) = 2*(Ny+1)*(Nz+1)*l+2*(Nz+1)*m+2*n+a
    for l in 0:Nx    
        for m in 0:Ny
            for n in 0:Nz
                for a in 1:2
                    for l_ in 0:Nx    
                        for m_ in 0:Ny
                            for n_ in 0:Nz
                                for a_ in 1:2
                                    sbra = s(l_,m_,n_,a_)
                                    sket = s(l,m,n,a)
                                    
                                    Opi = Op(l_,m_,n_,a_,l,m,n,a)
                                    if Opi!=0 || sbra==sket
                                        push!(Iv,sbra)
                                        push!(Jv,sket)
                                        push!(Opm,Opi)
                                    end
                                end
                            end
                        end
                    end
                end
            end
        end
    end
    return sparse(Iv,Jv,Opm)
end

### Our junction Hamiltonian is 
struct ParametersJ
    hw::Tuple{Float64,Float64,Float64}
    I::Tuple{Float64,Float64,Float64}
    phi::Tuple{Float64,Float64,Float64}
end

#divf(x,y) = y==0 ? 0 : x/y

const sigmax = [0 1;1 0] 
const sigmay = [0 -1im;1im 0]
const sigmaz = [1 0;0 -1] 

δ(m,n) = m==n ? 1 : 0

function H(l_,m_,n_,a_,l,m,n,a,p::ParametersJ)

    hw,I,phi= p.hw,p.I,p.phi

    sigma = (sigmax[a_,a],sigmay[a_,a],sigmaz[a_,a]) 
    bc = (sqrt(l+1)*δ(l_,l+1), sqrt(m+1)*δ(m_,m+1), sqrt(n+1)*δ(n_,n+1))
    ba = (sqrt(l)*δ(l_,l-1),sqrt(m)*δ(m_,m-1),sqrt(n)*δ(n_,n-1))

    E = δ(l_,l)*δ(m_,m)*δ(n_,n)

    s::Complex{Float64} = 0

    s += sum(hw.*(l+1/2,m+1/2,n+1/2).*E.*δ(a_,a))
    s += sum(sigma.*I.*(bc.+ba).*(δ(m_,m)*δ(n_,n),δ(l_,l)*δ(n_,n),δ(l_,l)*δ(m_,m)))
    s += sum(sigma.*I.*phi)*E
    return s
end

"""The mapping from enumeration to actual points"""
x(i,N,L) = i<= N//2 ?  -(L+1)^((N/2-i+1)/(N/2+1)) + 1/2*(1 + (L+1)^(1/(N/2 + 1))) : (L+1)^((i-N/2)/(N/2+1)) - 1/2*(1 + (L+1)^(1/(N/2 + 1)))

speye(n) = sparse(1.0I, n, n)

### Emulation of Mathematica function
function SparseArray(t::Type,f::Function,indicies::Tuple{Int,Int})
    n,m = indicies

    V = Array{t}(undef,0)
    I_ = Array{Int}(undef,0)
    J = Array{Int}(undef,0)

    for i in 1:n
        for j in 1:m
            val = f(i,j)
            if val!=0
                push!(I_,i)
                push!(J,j)
                push!(V,val)
            end
        end
    end

    return sparse(I_,J,V)
end

"""
Operator elements for first derivative on the 5 point grid
"""
function d1(i,j,y)
    x(p) = y(i+p) - y(i)

    j==i-2 ? -((x(-1)*x(1)*x(2))/(x(-2)*(x(-2)-x(-1))*(x(-2)-x(1))*(x(-2)-x(2)))) :
        j==i-1 ? (x(-2)*x(1)*x(2))/((x(-2)-x(-1))*x(-1)*(x(-1)-x(1))*(x(-1)-x(2))) :
        j==i ? -(1/x(-1))-1/x(1)-1/x(2)-1/x(-2) :
        j==i+1 ? (x(-2)*x(-1)*x(2))/((x(-2)-x(1))*x(1)*(x(1)-x(-1))*(x(1)-x(2))) :
        j==i+2 ? (x(-2)*x(-1)*x(1))/((x(-2)-x(2))*x(2)*(x(2)-x(-1))*(x(2)-x(1))) :
        0
end

"""
Operator elements for second derivative on the 5 point grid
"""
function d2(i,j,y)
    x(p) = y(i+p) - y(i)

    j==i-2 ? (2*(x(1)*x(2)+x(-1)*(x(1)+x(2))))/(x(-2)*(x(-2)-x(-1))*(x(-2)-x(1))*(x(-2)-x(2))) : 
        j==i-1 ? -((2*(x(1)*x(2)+x(-2)*(x(1)+x(2))))/((x(-2)-x(-1))*x(-1)*(x(-1)-x(1))*(x(-1)-x(2)))) :
        j==i ? (2*(x(1)*x(2)+x(-1)*(x(1)+x(2))+x(-2)*(x(-1)+x(1)+x(2))))/(x(-2)*x(-1)*x(1)*x(2)) :
        j==i+1 ? -((2*(x(-1)*x(2)+x(-2)*(x(-1)+x(2))))/((x(-2)-x(1))*x(1)*(x(1)-x(-1))*(x(1)-x(2)))) :
        j==i+2 ? -((2*(x(-1)*x(1)+x(-2)*(x(-1)+x(1))))/((x(-2)-x(2))*x(2)*(x(2)-x(-1))*(x(2)-x(1)))) :
        0
end

function ExcitonHamiltonian(α,(vx,vy),(mx,my),(p0x,p0y),L,N)
    
    sigmax = [0 1;1 0] 
    sigmay = [0 -1im;1im 0]
    sigmaz = [1 0;0 -1]
    
    ### Laplacian operator
    D2 = SparseArray(Float64,(i,j)->d2(i,j,x_->x(x_,N,L)),(N,N))
    Lap = kron(D2,speye(N))/mx + kron(speye(N),D2)/my

    ### Potential
    V = Float64[]
    for i in 1:N
        for j in 1:N
            x_ = x(i,N,L)
            y_ = x(j,N,L)
            r = sqrt(x_^2+y_^2)
            push!(V,-α/r)
        end
    end
    V = sparse(Diagonal(V))

    D1 = SparseArray(Float64,(i,j)->d1(i,j,x_->x(x_,N,L)),(N,N))
    Hm = kron(-1/2*Lap + V,speye(2)) + vx*kron(1im*kron(D1,speye(N)) + p0x*speye(N^2),sigmax) + vy*kron(1im*kron(speye(N),D1) + p0y*speye(N^2),sigmay)
end


function energies(α,v,m,p0,L,N,NEV,σoff)

    σ = σoff - maximum(v)^2/2 - maximum(v)*maximum(p0)

    println("p0 = $p0")
    run(`hostname`)
    BLAS.set_num_threads(1)

    Hm = ExcitonHamiltonian(α,v,m,p0,L,N)
    E = eigs(Hm,nev=NEV,tol=1e-6,maxiter=10000,which=:LM,sigma=σ)[1]
    E = sort(real(E))

    return E
end


function test()
    @show Hm = ExcitonHamiltonian(1,(0,0),(1,1),(0,0),10,50)
    @show eigs(Hm,nev=4,tol=1e-6,maxiter=10000,which=:SR)[1]
end
